# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard library imports
import os
import re
import sys
from argparse import ArgumentParser, RawTextHelpFormatter
from configparser import ConfigParser
from datetime import timedelta
from logging import Formatter, StreamHandler, getLogger
from pathlib import Path
from subprocess import check_call
from textwrap import dedent
from time import time

# Third-party imports
from git import Repo, rmtree
from ruamel.yaml import YAML
from setuptools_scm import get_version

fmt = Formatter('%(asctime)s [%(levelname)s] [%(name)s] -> %(message)s')
h = StreamHandler()
h.setFormatter(fmt)
logger = getLogger('BuildCondaPkgs')
logger.addHandler(h)
logger.setLevel('INFO')

HERE = Path(__file__).parent
BUILD = HERE / "build"
RESOURCES = HERE / "resources"
EXTDEPS = HERE.parent / "external-deps"
SPECS = BUILD / "specs.yaml"
REQUIREMENTS = HERE.parent / "requirements"
REQ_MAIN = REQUIREMENTS / 'main.yml'
REQ_WINDOWS = REQUIREMENTS / 'windows.yml'
REQ_MAC = REQUIREMENTS / 'macos.yml'
REQ_LINUX = REQUIREMENTS / 'linux.yml'

BUILD.mkdir(exist_ok=True)
SPYPATCHFILE = BUILD / "installers-conda.patch"

yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)


class BuildCondaPkg:
    """Base class for building a conda package for conda-based installer"""
    name = None
    norm = True
    source = None
    feedstock = None
    feedstock_branch = None
    shallow_ver = None

    def __init__(self, data={}, debug=False, shallow=False):
        self.logger = getLogger(self.__class__.__name__)
        if not self.logger.handlers:
            self.logger.addHandler(h)
        self.logger.setLevel('INFO')

        self.debug = debug

        self._bld_src = BUILD / self.name
        self._fdstk_path = BUILD / self.feedstock.split("/")[-1]
        self._patchfile = self._fdstk_path / "recipe" / "version.patch"

        self._get_source(shallow=shallow)
        self._get_version()

        self.data = {'version': self.version}
        self.data.update(data)

        self.recipe_append = {}
        self.recipe_clobber = {}

        self._recipe_patched = False

    def _get_source(self, shallow=False):
        """Clone source and feedstock to distribution directory for building"""
        self._build_cleanup()

        if self.source == HERE.parent:
            self._bld_src = self.source
            self.repo = Repo(self.source)
        else:
            # Determine source and commit
            if self.source is not None:
                remote = self.source
                commit = 'HEAD'
            else:
                cfg = ConfigParser()
                cfg.read(EXTDEPS / self.name / '.gitrepo')
                remote = cfg['subrepo']['remote']
                commit = cfg['subrepo']['commit']

            # Clone from source
            kwargs = dict(to_path=self._bld_src)
            if shallow:
                kwargs.update(shallow_exclude=self.shallow_ver)
                self.logger.info(
                    f"Cloning source shallow from tag {self.shallow_ver}...")
            else:
                self.logger.info("Cloning source...")
            self.repo = Repo.clone_from(remote, **kwargs)
            self.repo.git.checkout(commit)

        # Clone feedstock
        self.logger.info("Cloning feedstock...")
        kwargs = dict(to_path=self._fdstk_path)
        if self.feedstock_branch:
            kwargs.update(branch=self.feedstock_branch)
        Repo.clone_from(self.feedstock, **kwargs)

    def _build_cleanup(self):
        """Remove cloned source and feedstock repositories"""
        for src in [self._bld_src, self._fdstk_path]:
            if src.exists() and src != HERE.parent:
                logger.info(f"Removing {src}...")
                rmtree(src)

    def _get_version(self):
        """Get source version using setuptools_scm"""
        v = get_version(self._bld_src, normalize=self.norm)
        self.version = v.lstrip('v').split('+')[0]

    def _patch_source(self):
        pass

    def _patch_conda_build_config(self):
        file = self._fdstk_path / "recipe" / "conda_build_config.yaml"
        if not file.exists():
            return

        contents = yaml.load(file.read_text())
        file.rename(file.parent / ("_" + file.name))  # copy of original

        pyver = sys.version_info
        contents['python'] = [f"{pyver.major}.{pyver.minor}.* *_cpython"]
        yaml.dump(contents, file)

        self.logger.info(
            f"Patched 'conda_build_config.yaml' contents:\n{file.read_text()}"
        )

    def _patch_meta(self):
        file = self._fdstk_path / "recipe" / "meta.yaml"
        meta = file.read_text()

        # Replace jinja variable values
        for k, v in self.data.items():
            meta = re.sub(f".*set {k} =.*", f'{{% set {k} = "{v}" %}}', meta)

        file.rename(file.parent / ("_" + file.name))  # keep copy of original
        file.write_text(meta)

        self.logger.info(f"Patched 'meta.yaml' contents:\n{file.read_text()}")

    def _add_recipe_append(self):
        if self._patchfile.exists():
            self.recipe_append.update(
                {"source": {"patches": [self._patchfile.name]}}
            )

        if self.recipe_append:
            file = self._fdstk_path / "recipe" / "recipe_append.yaml"
            yaml.dump(self.recipe_append, file)
            self.logger.info(
                f"'recipe_append.yaml' contents:\n{file.read_text()}"
            )
        else:
            self.logger.info("Skipping 'recipe_append.yaml'.")

    def _add_recipe_clobber(self):
        self.recipe_clobber.update({
            "source": {
                "url": None,
                "sha256": None,
                "path": self._bld_src.as_posix()},
        })

        if self.recipe_clobber:
            file = self._fdstk_path / "recipe" / "recipe_clobber.yaml"
            yaml.dump(self.recipe_clobber, file)
            self.logger.info(
                f"'recipe_clobber.yaml' contents:\n{file.read_text()}"
            )
        else:
            self.logger.info("Skipping 'recipe_clobber.yaml'.")

    def patch_recipe(self):
        """
        Patch conda build recipe

        1. Patch conda_build_config.yaml
        2. Patch meta.yaml
        3. Add recipe_append.yaml
        4. Add recipe_clobber.yaml
        """
        if self._recipe_patched:
            return

        self._patch_conda_build_config()
        self._patch_meta()
        self._add_recipe_append()
        self._add_recipe_clobber()

        self._recipe_patched = True

    def build(self):
        """
        Build the conda package.

        1. Patch source
        2. Patch the recipe
        3. Build the package
        4. Remove cloned repositories
        """
        t0 = time()
        try:
            self._patch_source()
            self.patch_recipe()

            self.logger.info("Building conda package "
                             f"{self.name}={self.version}...")
            check_call([
                "conda", "build",
                "--skip-existing", "--build-id-pat={n}",
                "--no-test", "--no-anaconda-upload",
                str(self._fdstk_path / "recipe")
            ])
        finally:
            self._recipe_patched = False
            if self.debug:
                self.logger.info("Keeping cloned source and feedstock")
            else:
                self._build_cleanup()

            elapse = timedelta(seconds=int(time() - t0))
            self.logger.info(f"Build time = {elapse}")


class SpyderCondaPkg(BuildCondaPkg):
    name = "spyder"
    norm = False
    source = os.environ.get('SPYDER_SOURCE', HERE.parent)
    feedstock = "https://github.com/conda-forge/spyder-feedstock"
    feedstock_branch = "main"
    shallow_ver = "v5.3.2"

    def _patch_source(self):
        self.logger.info("Patching Spyder source...")

        file = self._bld_src / "spyder/__init__.py"
        file_text = file.read_text()
        ver_str = tuple(self.version.split('.'))
        file_text = re.sub(
            r'^(version_info = ).*',
            rf'\g<1>{ver_str}',
            file_text,
            flags=re.MULTILINE
        )
        file.write_text(file_text)

        # Only write patch if necessary
        if self.repo.git.diff():
            self.logger.info(f"Creating {self._patchfile.name}...")
            self.repo.git.diff(output=self._patchfile.as_posix())
            self.repo.git.stash()

    def patch_recipe(self):
        # Get current Spyder requirements
        spyder_base_reqs = ['python']
        spyder_base_reqs += yaml.load(
            REQ_MAIN.read_text())['dependencies']
        if os.name == 'nt':
            win_requirements =  yaml.load(
                REQ_WINDOWS.read_text())['dependencies']
            spyder_base_reqs += win_requirements
            spyder_base_reqs.append('ptyprocess >=0.5')
        elif sys.platform == 'darwin':
            mac_requirements =  yaml.load(
                REQ_MAC.read_text())['dependencies']
            if 'python.app' in mac_requirements:
                mac_requirements.remove('python.app')
            spyder_base_reqs += mac_requirements
            spyder_base_reqs.append('__osx')
        else:
            linux_requirements = yaml.load(
                REQ_LINUX.read_text())['dependencies']
            spyder_base_reqs += linux_requirements
            spyder_base_reqs.append('__linux')

        spyder_reqs = [f"spyder-base =={self.version}"]
        for req in spyder_base_reqs.copy():
            if req.startswith(('pyqt ', 'pyqtwebengine ', 'qtconsole ')):
                spyder_reqs.append(req)
                spyder_base_reqs.remove(req)

            if req.startswith('qtconsole '):
                spyder_base_reqs.append(
                    req.replace('qtconsole', 'qtconsole-base')
                )

        self.recipe_clobber.update({
            "requirements": {"run": spyder_base_reqs},
            # Since outputs is a list, the entire list must be reproduced with
            # the current run requirements
            "outputs": [
                {
                    "name": "spyder-base"
                },
                {
                    "name": "spyder",
                    "build": {"noarch": "python"},
                    "requirements": {"run": spyder_reqs},
                    "test": {
                        "requires": ["pip"],
                        "commands": ["spyder -h", "python -m pip check"],
                        "imports": ["spyder"]
                    }
                }
            ]
        })

        super().patch_recipe()


class PylspCondaPkg(BuildCondaPkg):
    name = "python-lsp-server"
    source = os.environ.get('PYTHON_LSP_SERVER_SOURCE')
    feedstock = "https://github.com/conda-forge/python-lsp-server-feedstock"
    feedstock_branch = "main"
    shallow_ver = "v1.4.1"


class QtconsoleCondaPkg(BuildCondaPkg):
    name = "qtconsole"
    source = os.environ.get('QTCONSOLE_SOURCE')
    feedstock = "https://github.com/conda-forge/qtconsole-feedstock"
    feedstock_branch = "main"
    shallow_ver = "5.3.1"


class SpyderKernelsCondaPkg(BuildCondaPkg):
    name = "spyder-kernels"
    source = os.environ.get('SPYDER_KERNELS_SOURCE')
    feedstock = "https://github.com/conda-forge/spyder-kernels-feedstock"
    feedstock_branch = "rc"
    shallow_ver = "v2.3.1"


PKGS = {
    SpyderCondaPkg.name: SpyderCondaPkg,
    PylspCondaPkg.name: PylspCondaPkg,
    QtconsoleCondaPkg.name: QtconsoleCondaPkg,
    SpyderKernelsCondaPkg.name: SpyderKernelsCondaPkg
}

if __name__ == "__main__":
    p = ArgumentParser(
        description=dedent(
            """
            Build conda packages to local channel.

            This module builds conda packages for Spyder and external-deps for
            inclusion in the conda-based installer. The following classes are
            provided for each package:
                SpyderCondaPkg
                PylspCondaPkg
                QdarkstyleCondaPkg
                QtconsoleCondaPkg
                SpyderKernelsCondaPkg

            Spyder will be packaged from this repository (in its checked-out
            state). qtconsole, spyder-kernels, and python-lsp-server will be
            packaged from the remote and commit specified in their respective
            .gitrepo files in external-deps.

            Alternatively, any external-deps may be packaged from an arbitrary
            git repository (in its checked out state) by setting the
            appropriate environment variable from the following:
                SPYDER_SOURCE
                PYTHON_LSP_SERVER_SOURCE
                QDARKSTYLE_SOURCE
                QTCONSOLE_SOURCE
                SPYDER_KERNELS_SOURCE
            """
        ),
        usage="python build_conda_pkgs.py "
              "[--build BUILD [BUILD] ...] [--debug] [--shallow]",
        formatter_class=RawTextHelpFormatter
    )
    p.add_argument(
        '--debug', action='store_true', default=False,
        help="Do not remove cloned sources and feedstocks"
    )
    p.add_argument(
        '--build', nargs="+", default=PKGS.keys(),
        help=("Space-separated list of packages to build. "
              f"Default is {list(PKGS.keys())}")
    )
    p.add_argument(
        '--shallow', action='store_true', default=False,
        help="Perform shallow clone for build")
    args = p.parse_args()

    logger.info(f"Building local conda packages {list(args.build)}...")
    t0 = time()

    if SPECS.exists():
        specs = yaml.load(SPECS.read_text())
    else:
        specs = {k: "" for k in PKGS}

    for k in args.build:
        pkg = PKGS[k](debug=args.debug, shallow=args.shallow)
        pkg.build()
        specs[k] = "=" + pkg.version

    yaml.dump(specs, SPECS)

    elapse = timedelta(seconds=int(time() - t0))
    logger.info(f"Total build time = {elapse}")



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard library imports
from ast import literal_eval
import os.path as osp

# Third party imports
import pytest
import yaml

# Local imports
from spyder.dependencies import DESCRIPTIONS, OPTIONAL, PY38

# Constants
HERE = osp.abspath(osp.dirname(__file__))
ROOT_PATH = osp.dirname(osp.dirname(HERE))
ENV_FPATH = osp.join(ROOT_PATH, 'binder', 'environment.yml')
REQ_FPATH = osp.join(ROOT_PATH, 'requirements', 'main.yml')
REQ_WINDOWS_FPATH = osp.join(ROOT_PATH, 'requirements', 'windows.yml')
REQ_MAC_FPATH = osp.join(ROOT_PATH, 'requirements', 'macos.yml')
REQ_LINUX_FPATH = osp.join(ROOT_PATH, 'requirements', 'linux.yml')
REQ_TEST_FPATH = osp.join(ROOT_PATH, 'requirements', 'tests.yml')
SETUP_FPATH = osp.join(ROOT_PATH, 'setup.py')


def parse_environment_yaml(fpath):
    """
    Parse an environment yaml file and return a dict of deps and versions.
    """
    with open(fpath, 'r') as fh:
        data = yaml.load(fh, Loader=yaml.FullLoader)

    deps = {}
    yaml_deps = data.get('dependencies')
    for dep in yaml_deps:
        if isinstance(dep, dict):
            continue
        elif dep == 'websockify':
            continue
        else:
            parts = dep.split(' ')
            if len(parts) > 1:
                ver = parts[-1]
                if ver[0] == '=':
                    ver = '=' + ver

                deps[parts[0]] = ver
            else:
                deps[parts[0]] = None

    return deps


def parse_spyder_dependencies():
    """
    Parse Spyder dependencies and return a dict of deps and versions.
    """
    deps = {}
    for dep in DESCRIPTIONS:
        if dep.get('kind', None) == OPTIONAL:
            continue

        ver = dep['required_version']
        if ver:
            if ';' in ver:
                ver = ver.replace(';', ',')
            elif ver[0] == '=':
                ver = '=' + ver

        deps[dep['package_name'].lower()] = ver

    return deps


def parse_setup_install_requires(fpath):
    """
    Parse Spyder setup.py and return a dict of deps and versions.
    """
    with open(fpath, 'r') as fh:
        data = fh.read()

    # Extract dependencies
    lines = data.split('\n')
    is_dep_line = False
    deps_list = []
    for line in lines:
        if is_dep_line:
            if line.strip() in ('],', ']'):
                # End of dependency lines
                is_dep_line = False
            elif not line.strip().startswith('#'):
                # Add dependency
                deps_list.append(line.strip().strip(',').strip("'"))
        if (
            line.strip() in ("'pyqt5': [", "install_requires += [")
            and not is_dep_line
        ):
            # Depencies begin on next line
            is_dep_line = True

    # Process dependencies
    deps = {}
    for dep in deps_list:
        dep = dep.split(';')[0]
        name, ver = None, None

        for sep in ['>=', '==', '<=', '<', '>']:
            if sep in dep:
                name, ver = dep.split(sep)
                name = name.split('[')[0]  # Discard e.g. [all]
                ver = sep + ver  # Include comparator
                break

        # Transform pypi to conda name
        if name == 'pyqt5':
            name = 'pyqt'

        deps[name] = ver

    return deps


def parse_setup_extra_requires(fpath):
    """
    Parse Spyder setup.py and return a dict of deps and versions.
    """
    deps = {}
    with open(fpath, 'r') as fh:
        data = fh.read()

    lines = data.split('\n')
    start = None
    end = None
    for idx, line in enumerate(lines):
        if line.startswith('extras_require = '):
            start = idx + 1

        if start is not None and line.startswith('}'):
            end = idx
            break

    dep_dict = literal_eval('{' + '\n'.join(lines[start:end + 1]))
    dep_list = dep_dict.get('test')
    dep_list = [item for item in dep_list if item[0] != '#']
    for dep in dep_list:
        dep = dep.split(';')[0]
        name, ver = None, None

        for sep in ['>=', '==', '<=', '<', '>']:
            if sep in dep:
                idx = dep.index(sep)
                name = dep[:idx]
                ver = dep[idx:]
                break

        if name is not None:
            name = name.split('[')[0]
        else:
            name = dep.split('[')[0]

        # Transform pypi to conda name
        if name == 'pyqt5':
            name = 'pyqt'

        deps[name] = ver
    print(deps)
    return deps


def test_dependencies_for_binder_in_sync():
    """
    Binder environment yaml should be the sum of main.yml and tests.yml
    requirements.
    """
    spyder_env = parse_environment_yaml(ENV_FPATH)
    main_reqs = parse_environment_yaml(REQ_FPATH)
    test_reqs = parse_environment_yaml(REQ_TEST_FPATH)
    linux_reqs = parse_environment_yaml(REQ_LINUX_FPATH)

    # Check that the requirement files match the environment yaml file
    full_reqs = {}
    full_reqs.update(main_reqs)
    full_reqs.update(linux_reqs)
    full_reqs.update(test_reqs)

    assert spyder_env == full_reqs


@pytest.mark.skipif(PY38, reason="Fails in Python 3.8")
def test_dependencies_for_spyder_dialog_in_sync():
    """
    Spyder dependencies dialog should share deps with main.yml.
    """
    spyder_deps = parse_spyder_dependencies()
    main_reqs = parse_environment_yaml(REQ_FPATH)
    windows_reqs = parse_environment_yaml(REQ_WINDOWS_FPATH)
    mac_reqs = parse_environment_yaml(REQ_MAC_FPATH)
    linux_reqs = parse_environment_yaml(REQ_LINUX_FPATH)

    full_reqs = {}
    full_reqs.update(main_reqs)
    full_reqs.update(windows_reqs)
    full_reqs.update(mac_reqs)
    full_reqs.update(linux_reqs)

    # These packages are not declared in our dependencies dialog
    for dep in ['pyqt', 'pyqtwebengine', 'python.app', 'fzf', 'fcitx-qt5']:
        full_reqs.pop(dep)

    assert spyder_deps == full_reqs


def test_dependencies_for_spyder_setup_install_requires_in_sync():
    """
    Spyder setup.py should share deps with main.yml.
    """
    spyder_setup = parse_setup_install_requires(SETUP_FPATH)
    main_reqs = parse_environment_yaml(REQ_FPATH)
    windows_reqs = parse_environment_yaml(REQ_WINDOWS_FPATH)
    mac_reqs = parse_environment_yaml(REQ_MAC_FPATH)
    linux_reqs = parse_environment_yaml(REQ_LINUX_FPATH)

    full_reqs = {}
    full_reqs.update(main_reqs)
    full_reqs.update(windows_reqs)
    full_reqs.update(mac_reqs)
    full_reqs.update(linux_reqs)

    # We can't declare these as dependencies in setup.py
    for dep in ['python.app', 'fzf', 'fcitx-qt5']:
        full_reqs.pop(dep)
    # Ignored `pyqt5-sip` constraint on conda
    spyder_setup.pop('pyqt5-sip')

    assert spyder_setup == full_reqs


def test_dependencies_for_spyder_setup_extras_requires_in_sync():
    """
    Spyder setup.py extra_requires should share deps with tests.yml.
    """
    spyder_extras_setup = parse_setup_extra_requires(SETUP_FPATH)
    spyder_test_reqs = parse_environment_yaml(REQ_TEST_FPATH)

    assert spyder_extras_setup == spyder_test_reqs



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tests for the preferences dialog.
"""

from spyder.plugins.variableexplorer.widgets.preferences import (
    PreferencesDialog
)


def test_preferences_float_format(qtbot):
    """
    Test that reading and setting float_format works.
    """
    dialog = PreferencesDialog('dataframe')
    dialog.float_format = 'old'
    assert dialog.format_input.text() == 'old'

    dialog.format_input.setText('new')
    assert dialog.float_format == 'new'


def test_preferences_background(qtbot):
    """
    Test default and varying background buttons.
    """
    dialog = PreferencesDialog('dataframe')
    dialog.varying_background = True

    assert dialog.varying_background
    assert not dialog.default_background_button.isChecked()
    assert dialog.varying_background_button.isChecked()
    assert dialog.global_button.isEnabled()
    assert dialog.by_column_button.isEnabled()

    dialog.default_background_button.toggle()

    assert not dialog.varying_background
    assert dialog.default_background_button.isChecked()
    assert not dialog.varying_background_button.isChecked()
    assert not dialog.global_button.isEnabled()
    assert not dialog.by_column_button.isEnabled()

    dialog.varying_background_button.toggle()

    assert dialog.varying_background
    assert not dialog.default_background_button.isChecked()
    assert dialog.varying_background_button.isChecked()
    assert dialog.global_button.isEnabled()
    assert dialog.by_column_button.isEnabled()


def test_preferences_coloring_algo(qtbot):
    """
    Test coloring algo buttons (global and by column)
    """
    dialog = PreferencesDialog('dataframe')
    dialog.varying_background = True
    dialog.global_algo = True

    assert dialog.global_algo
    assert dialog.global_button.isChecked()
    assert not dialog.by_column_button.isChecked()

    dialog.by_column_button.toggle()

    assert not dialog.global_algo
    assert not dialog.global_button.isChecked()
    assert dialog.by_column_button.isChecked()

    dialog.global_button.toggle()

    assert dialog.global_algo
    assert dialog.global_button.isChecked()
    assert not dialog.by_column_button.isChecked()


def test_preferences_for_arrays(qtbot):
    """
    Check that dialog for arrays does not have coloring algo buttons.
    """
    dialog = PreferencesDialog('array')
    assert 'global_button' not in dir(dialog)
    assert 'by_column_button' not in dir(dialog)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
"""Tests for the class and method panel selector."""

# Third party imports
from flaky import flaky
from qtpy.QtCore import QPoint, Qt
from qtpy.QtWidgets import QListView
import pytest


text = """
from collections import OrderedDict


class SomeObject:  # Block number 4

    def __init__(self):
        pass

    def hello_1(self):  # Block number 9
        pass

    def hello_2(self):
        pass


class SomeOtherObject:

    def __init__(self):
        pass

    def hello_3(self):
        pass

        def nested_func():
            pass
"""


@pytest.mark.order(2)
@flaky(max_runs=5)
def test_class_func_selector(completions_codeeditor, qtbot):
    code_editor, _ = completions_codeeditor
    panel = code_editor.classfuncdropdown
    panel.setVisible(True)
    code_editor.toggle_automatic_completions(False)
    code_editor.set_text(text)

    # Wait for symbols info to arrive
    qtbot.wait(3000)

    class_names = [item['name'] for item in panel.classes]
    func_names = [item['name'] for item in panel.funcs]

    assert len(panel.classes) == 2
    assert len(panel.funcs) == 6
    assert 'SomeObject' in class_names
    assert 'SomeOtherObject' in class_names
    assert '__init__' in func_names
    assert 'nested_func' in func_names

    # Check line is followed
    for _ in range(7):
        qtbot.keyPress(code_editor, Qt.Key_Down)
    assert panel.class_cb.currentText() == 'SomeObject'
    assert panel.method_cb.currentText() == 'SomeObject.__init__'

    for _ in range(18):
        qtbot.keyPress(code_editor, Qt.Key_Down)
    assert panel.class_cb.currentText() == 'SomeOtherObject'
    assert (panel.method_cb.currentText()
            == 'SomeOtherObject.hello_3.nested_func')

    # Check go to line works for class selector
    qtbot.mouseClick(panel.class_cb, Qt.LeftButton, pos=QPoint(5, 5))
    listview = panel.class_cb.findChild(QListView)
    qtbot.keyPress(listview, Qt.Key_Up)
    qtbot.keyPress(listview, Qt.Key_Return)
    qtbot.wait(1000)
    cursor = code_editor.textCursor()
    assert cursor.blockNumber() == 4
    assert panel.method_cb.currentIndex() == 0

    # Check go to line works for func selector
    panel.method_cb.setFocus()
    qtbot.mouseClick(panel.method_cb, Qt.LeftButton, pos=QPoint(5, 5))
    listview = panel.method_cb.findChild(QListView)
    qtbot.keyPress(listview, Qt.Key_Down)
    qtbot.keyPress(listview, Qt.Key_Down)
    qtbot.keyPress(listview, Qt.Key_Return)
    qtbot.wait(1000)
    cursor = code_editor.textCursor()
    assert cursor.blockNumber() == 9
    assert panel.class_cb.currentIndex() == 1

    panel.setVisible(False)
    code_editor.toggle_automatic_completions(True)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Tests for the Plots plugin.
"""

from unittest.mock import Mock

import pytest

from spyder.config.base import running_in_ci
from spyder.config.manager import CONF
from spyder.plugins.plots.plugin import Plots
from spyder.plugins.plots.widgets.main_widget import PlotsWidgetActions
from spyder.plugins.plots.widgets.tests.test_plots_widgets import (
    add_figures_to_browser,
)
from spyder.utils.stylesheet import APP_STYLESHEET


@pytest.fixture
def plots_plugin(qapp, qtbot):
    plots = Plots(None, configuration=CONF)
    plots.get_widget().setMinimumSize(700, 500)
    plots.get_widget().add_shellwidget(Mock())
    qtbot.addWidget(plots.get_widget())
    if not running_in_ci():
        qapp.setStyleSheet(str(APP_STYLESHEET))
    with qtbot.waitExposed(plots.get_widget()):
        plots.get_widget().show()

    return plots


def test_fit_action(plots_plugin, tmpdir, qtbot):
    """Test that the fit action works as expected."""
    main_widget = plots_plugin.get_widget()

    # Action should not be checked when no plots are available
    assert not main_widget.fit_action.isChecked()

    # Add some plots
    figbrowser = main_widget.current_widget()
    figviewer = figbrowser.figviewer
    add_figures_to_browser(figbrowser, 2, tmpdir)

    # Action should be checked now and zoom factor less than 100%
    assert main_widget.fit_action.isChecked()
    assert main_widget.zoom_disp.value() < 100

    # Plot should be zoomed in at full size when unchecking action
    main_widget.fit_action.setChecked(False)
    assert main_widget.zoom_disp.value() == 100
    qtbot.wait(200)
    assert figviewer.horizontalScrollBar().isVisible()

    # Action should be checked for next plot
    main_widget.next_plot()
    assert main_widget.fit_action.isChecked()

    # Action should be unchecked when going back to the previous plot
    main_widget.previous_plot()
    assert not main_widget.fit_action.isChecked()

    # Plot should be fitted when checking the action again
    main_widget.fit_action.setChecked(True)
    assert main_widget.zoom_disp.value() < 100
    qtbot.wait(200)
    assert not figviewer.horizontalScrollBar().isVisible()


def test_zoom_actions(plots_plugin, qtbot, tmpdir):
    """Test that the behavior of the zoom actions work as expected."""
    main_widget = plots_plugin.get_widget()
    zoom_in_action = main_widget.get_action(PlotsWidgetActions.ZoomIn)
    zoom_out_action = main_widget.get_action(PlotsWidgetActions.ZoomOut)

    # Zoom in/out actions should be disabled when no plots are available
    assert not zoom_in_action.isEnabled()
    assert not zoom_out_action.isEnabled()

    # Add some plots
    figbrowser = main_widget.current_widget()
    figviewer = figbrowser.figviewer
    add_figures_to_browser(figbrowser, 3, tmpdir)

    # Zoom in/out actions should be enabled now
    assert zoom_in_action.isEnabled()
    assert zoom_out_action.isEnabled()

    # Zoom in first plot twice
    for __ in range(2):
        main_widget.zoom_in()
        qtbot.wait(100)

    # Save zoom and scrollbar values to test them later
    qtbot.wait(200)
    zoom_1 = main_widget.zoom_disp.value()
    vscrollbar = figviewer.verticalScrollBar()
    hscrollbar = figviewer.horizontalScrollBar()
    vscrollbar_value_1 = vscrollbar.value()
    hscrollbar_value_1 = hscrollbar.value()

    # Fit action should be unchecked now
    assert not main_widget.fit_action.isChecked()

    # Next plot should be still fitted
    main_widget.next_plot()
    assert main_widget.fit_action.isChecked()
    assert main_widget.zoom_disp.value() < 100
    assert not hscrollbar.isVisible()
    assert not vscrollbar.isVisible()

    # Zoom out twice this plot
    for __ in range(2):
        main_widget.zoom_out()
        qtbot.wait(100)

    # Fit action should be unchecked now
    assert not main_widget.fit_action.isChecked()

    # Save zoom level for later
    zoom_2 = main_widget.zoom_disp.value()

    # Next plot should be still fitted
    main_widget.next_plot()
    assert main_widget.fit_action.isChecked()

    # Return to the first plot
    for __ in range(2):
        main_widget.previous_plot()
        qtbot.wait(100)

    # Check zoom level and scrollbars are restored
    qtbot.wait(200)
    assert main_widget.zoom_disp.value() == zoom_1
    assert vscrollbar.value() == vscrollbar_value_1
    assert hscrollbar.value() == hscrollbar_value_1

    # Move to next plot and check zoom level is restored
    main_widget.next_plot()
    assert main_widget.zoom_disp.value() == zoom_2



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2013-2016 Colin Duquesnoy and others (see pyqode/AUTHORS.rst)
# Copyright (c) 2016- Spyder Project Contributors (see AUTHORS.txt)
#
# Distributed under the terms of the MIT License
# (see NOTICE.txt in the Spyder root directory for details)
# -----------------------------------------------------------------------------

"""
This module contains the Manager API.

Adapted from pyqode/core/api/manager.py of the
`PyQode project <https://github.com/pyQode/pyQode>`_.
Original file:
<https://github.com/pyQode/pyqode.core/blob/master/pyqode/core/api/manager.py>
"""

# Standard library imports
import weakref


class Manager(object):
    """
    A manager manages a specific aspect of a CodeEditor instance:
        - panels management and drawing

    Managers are typically created internally when you create a CodeEditor.
    You interact with them later.

    ::
        editor = CodeEditor()

        # use the panels controller to install a panel
        editor.panels.install(MyPanel(), MyPanel.Position.Right)
        my_panel = editor.panels.get(MyPanel)

        # and so on

    """

    @property
    def editor(self):
        """Return a reference to the parent CodeEditor widget."""
        return self._editor()

    def __init__(self, editor):
        """:param editor: CodeEditor instance to control."""
        super().__init__()
        self._editor = weakref.ref(editor)
